{% block head %}
    <link rel="stylesheet" href="/static/shows.css">
    <title>{% block title %}{% endblock title %}</title>
{% endblock head %}
{% block content %}
    <div class="row">
        <div class="col-md-12 container-fluid">
            <h1 class="header">Listing Shows</h1>
            {% for show in shows %}
                <div class="col-3">
                    <p>{{ show.name }} ({{ show.startyear }})</p>
                    <p>tvdb: {{ show.indexerid }}</p>
                    <p>imdb: {{ show.imdb_id }}</p>
                </div>
            {% endfor %}
        </div>
    </div>
{% endblock content %}
{% block footer %}
    <script src="/static/shows.js"></script>
{% endblock footer %}
